/*
   fastmemcpybench.c used to benchmark fastmemcpy.h code from libvo.
     
   Note: this code can not be used on PentMMX-PII because they contain
   a bug in rdtsc. For Intel processors since P6(PII) rdpmc should be used
   instead. For PIII it's disputable and seems bug was fixed but I don't
   tested it.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <inttypes.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/time.h>

#define MSGT_CLASS MSGT_GLOBAL
#include "../config.h"
#include "../mplayerxp/__mp_msg.h"
#include "../mplayerxp/cpudetect.h"
#include "../mplayerxp/libvo/fastmemcpy.h"


#if defined( ARCH_X86 ) || defined(ARCH_X86_64)
#define CAN_COMPILE_X86_ASM
#endif

#ifdef CAN_COMPILE_X86_ASM

#undef HAVE_MMX
#undef HAVE_MMX2
#undef HAVE_3DNOW
#undef HAVE_3DNOW2
#undef HAVE_SSE
#undef HAVE_SSE2
#undef HAVE_SSE3
#undef HAVE_SSSE3
#undef HAVE_SSE4
#undef HAVE_SSE5
/*MMX versions*/
#ifdef CAN_COMPILE_MMX
#undef RENAME
#define HAVE_MMX
#undef HAVE_MMX2
#undef HAVE_3DNOW
#define RENAME(a) a ## _MMX
#include "asmopt_template.h"
#endif

/*MMX2 versions*/
#ifdef CAN_COMPILE_MMX2
#undef RENAME
#define HAVE_MMX
#define HAVE_MMX2
#undef HAVE_3DNOW
#define RENAME(a) a ## _MMX2
#include "asmopt_template.h"
#endif

/*3DNOW versions*/
#ifdef CAN_COMPILE_3DNOW
#undef RENAME
#define HAVE_MMX
#undef HAVE_MMX2
#define HAVE_3DNOW
#define RENAME(a) a ## _3DNow
#include "asmopt_template.h"
#endif

/*3DNOW versions*/
#ifdef CAN_COMPILE_3DNOW2
#undef RENAME
#define HAVE_MMX
#define HAVE_MMX2
#define HAVE_3DNOW
#define HAVE_3DNOW2
#define RENAME(a) a ## _3DNow2
#include "asmopt_template.h"
#endif

#undef HAVE_3DNOW
#undef HAVE_3DNOWEX
/*SSE versions*/
#ifdef CAN_COMPILE_SSE
#undef RENAME
#define HAVE_MMX
#define HAVE_MMX2
#define HAVE_SSE
#define RENAME(a) a ## _SSE
#include "asmopt_template.h"
#endif

/*SSE2 versions*/
#ifdef CAN_COMPILE_SSE2
#undef RENAME
#define HAVE_MMX
#define HAVE_MMX2
#define HAVE_SSE
#define HAVE_SSE2
#define RENAME(a) a ## _SSE2
#include "asmopt_template.h"
#endif

#endif 

/* generic version */
#undef RENAME
#undef HAVE_MMX
#undef HAVE_MMX2
#undef HAVE_3DNOW
#define RENAME(a) a ## _C
#include "asmopt_template.h"

#define ARR_SIZE (1024*64*2)
unsigned verbose=1;
extern CpuCaps gCpuCaps;

/* Fixme: put here any complexness of source array filling */
#define INIT_ARRAYS(x) \
{\
	for(i=0; i<x; i++) srca[i] = i; \
	for(i=0; i<x; i++) dsta[i] = (~i)+128; \
}

// Returns current time in microseconds
unsigned int GetTimer(){
  struct timeval tv;
  struct timezone tz;
//  float s;
  gettimeofday(&tv,&tz);
//  s=tv.tv_usec;s*=0.000001;s+=tv.tv_sec;
  return (tv.tv_sec*1000000+tv.tv_usec);
}  

static inline unsigned long long int read_tsc( void )
{
  unsigned long long int retval;
  __asm __volatile ("rdtsc":"=A"(retval)::"memory");
  return retval;
}

unsigned char  __attribute__((aligned(4096)))srca[ARR_SIZE],dsta[ARR_SIZE];

int main( void )
{
  unsigned long long int v1,v2;
  int i;
  unsigned int t;
  mp_msg_init(verbose);
  GetCpuCaps(&gCpuCaps);
  MSG_INFO("CPUflags: MMX: %d MMX2: %d 3DNow: %d 3DNow2: %d SSE: %d SSE2: %d\n",
      gCpuCaps.hasMMX,gCpuCaps.hasMMX2,
      gCpuCaps.has3DNow, gCpuCaps.has3DNowExt,
      gCpuCaps.hasSSE, gCpuCaps.hasSSE2);


#ifdef CAN_COMPILE_X86_ASM
// ordered per speed fasterst first
#ifdef CAN_COMPILE_SSE2
    if(gCpuCaps.hasSSE2)
    {
	FILE *out=fopen("asmopt.sse2","wb");
	MSG_V("Using SSE2 Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_SSE2(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("SSE2: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#ifdef CAN_COMPILE_SSE
    if(gCpuCaps.hasSSE)
    {
	FILE *out=fopen("asmopt.sse","wb");
	MSG_V("Using SSE Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_SSE(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("SSE: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#ifdef CAN_COMPILE_3DNOW2
    if(gCpuCaps.has3DNowExt)
    {
	FILE *out=fopen("asmopt.3dnow2","wb");
	MSG_V("Using 3DNow2 Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_3DNow(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("3DNow2: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#ifdef CAN_COMPILE_3DNOW
    if(gCpuCaps.has3DNow)
    {
	FILE *out=fopen("asmopt.3dnow","wb");
	MSG_V("Using 3DNow Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_3DNow(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("3DNow: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#ifdef CAN_COMPILE_MMX2
    if(gCpuCaps.hasMMX2)
    {
	FILE *out=fopen("asmopt.mmx2","wb");
	MSG_V("Using MMX2 Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_MMX2(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("MMX2: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#ifdef CAN_COMPILE_MMX
    if(gCpuCaps.hasMMX)
    {
	FILE *out=fopen("asmopt.mmx","wb");
	MSG_V("Using MMX Optimized version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_MMX(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("MMX: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
#endif
#endif
    {
	FILE *out=fopen("asmopt.gen","wb");
	MSG_V("Using generic version\n");
	t=GetTimer();
	v1 = read_tsc();
	INIT_ARRAYS(ARR_SIZE);
	convert_C(dsta,srca,ARR_SIZE);
	v2 = read_tsc();
	t=GetTimer()-t;
	MSG_INFO("GEN: cpu clocks=%llu = %dus\n",v2-v1,t);
	if(out) fwrite(dsta,ARR_SIZE,1,out);
	fclose(out);
    }
  return 0;
}
